#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/shm.h>
#include <sys/param.h>
#include <assert.h>

#include "config_unix.h"
#include "config_win32.h"
#include "net_udp.h"
#include "debug.h"

#include "util.h"
#include "config.h"
#include "equation.h"
#include "lossfraction.h"

#define LOSS_V2
#undef LOSS_V3

typedef struct buffer_for_data {
	u_int32 packet_id;
	char data[DATA_SIZE];
	int data_size;
	int end_flag;
	struct buffer_for_data *next;
} data_buffer;

struct shared {
	int end_receiver;
	u_int32 sendpkt;
};

int shm_init(struct shared *info) {
	info->end_receiver = 0;
	info->sendpkt = 0;
	return 0;
}

int	ppl_cur;
int	ppl_hist[HIST_SIZE];
int	ts_hist[HIST_SIZE];	/* in ms */
int	seq_hist[HIST_SIZE];

u_int32	round;
int	num_loss;
int32	mseq_offs, high_offs;
int32	ppl_start, ppl_end;

int	weights;
int	const_weights;
float	weight[MAX_WEIGHTS];

int	ppl[MAX_WEIGHTS + 1];
float	new_mult; /* deweighting factor */

int	lost_packet_in_intervall[8];

double	(*lossfraction)();

void init_globals() {
	int i;
	for (i = 0; i < HIST_SIZE; ++i) {
		ppl_hist[i] = 0;
		ts_hist[i] = 0;
		seq_hist[i] = 0;
	}
	for (i = 0; i <= MAX_WEIGHTS; ++i) {
		ppl[i] = 0;
	}
	mseq_offs = 0;
	high_offs = -1;
	ppl_start = ppl_end = 0;
	ppl_cur = 0;
	weights = 8;
	round = 0;
	num_loss = 0;
	const_weights = 0;
	new_mult = 1.0;

	for(i = 0; i < 7; i++) {
		lost_packet_in_intervall[i]=0;
	}

#ifdef LOSS_V2
	lossfraction = lossfraction_v2;
#else
#ifdef LOSS_V3
	lossfraction = lossfraction_v3;
#else
	lossfraction = lossfraction_v1;
#endif
#endif
}


void set_weights() {
	int i;

	if (const_weights) {
		for (i = 0; i < weights; ++i)
			weight[i] = 1.00;
	} else {
		double delta = 1.0 / ((weights + 1) / 2 + 1);
		for (i = 0; i < weights / 2; ++i)
			weight[i] = 1.00;
		for (i = weights / 2; i < weights; ++i)
			weight[i] = weight[i-1] - delta;
	}
}


void log_loss(int64 arrival) {
	int i;
	double l_fake;

	/* log params */
	ppl_hist[num_loss - 1] = ppl[ppl_cur];
	ts_hist[num_loss - 1] = arrival / MS_TO_US;
	seq_hist[num_loss - 1] = high_offs + 1;


	#ifdef LOSS_V2
	if (new_mult < 1.0) {
		/* fake history */

		l_fake = lossfraction();
		new_mult = 1.0; /* reset deweighting factor */
		for (i = 0; i <= MAX_WEIGHTS; ++i) {
			if (ppl[i] > 0) {/* only fake intervals that were already used */
				ppl[i] = (int) (1.0 / l_fake + 0.5);
			}
		}
	}
#endif

	ppl_cur = (ppl_cur + 1) % (weights + 1);
	ppl[ppl_cur] = 0;
	ppl[ppl_cur] += mseq_offs - high_offs - 1; /* account for missing packets */
}

/* 
The mulTFRC equation needs other variables for the equation
n  ... number of TFRC flows
pe ... loss event probability of the cumulative flow
b  ... number of packets acknowledged by one packet - is 1
RTT... round trip time
T  ... initial value for retransmission timer - defined as 3
j  ... average number of packets lost in a loss event
*/
double mulTFRC_equation(double n, double pe, double b, double RTT, double T, double j) {
	double j1 = j;
	if(n<12)j1=n-n*pow((n-1)/n,j1);
	if (j1 > n)
		j1 = ceil(n);
	double a = sqrt(pe * b * j1 * (24 * n * n + pe * b * j1 * (n * n - 4  * n * j1 + 4 * j1 * j1)));
	double x = (j1 * pe * b * (2 * j1 - n) + a) / (6 * n * n * pe);
	double w = n * x / (2 * b ) * (1 + 3 * n / j1);
	double z = T * (1 + 32 * pe * pe) / (1 - pe);
	double q1 = j * n / w;
	if (q1 > n)
		q1 = n;
	double q;
	if ((q1 * z / (x * RTT)) >= n) {
		q = n;
	} else {
		q = q1 * z / (x * RTT);
	}
	return (((1 - q / n) / (pe * x * RTT) + q / (z * (1 - pe)))*PKT_SIZE);
}

double weight_lost_packets() {
	double weights[8] = {1.0, 1.0, 1.0, 1.0, 0.8, 0.6, 0.4, 0.2};
	double sum = 0;
	double sum_weights = 0;
	double avg_lost_packets = 0;
	int i;

	for (i = 0; i < 8; i++) {
		if (lost_packet_in_intervall[i]>0){
			sum = sum + ( weights[i] * lost_packet_in_intervall[i] ); 
			sum_weights = sum_weights + weights[i];	
		}
	}
	avg_lost_packets = sum / sum_weights;
	return avg_lost_packets;
}

int update_history(u_int32 mseq, int64 arrival, double rtt) {
	static int64 lastloss = 0;
	int i;
	mseq_offs = mseq;

	if (mseq_offs > high_offs + 1) { /* new loss */
		if ((arrival - lastloss > rtt * MS_TO_US)) { /* new loss event */
			++num_loss;
			if (num_loss >= HIST_SIZE) {
				fprintf(stderr, "loss history buffer overflow\n");
				return 0;
			}
			/* shift lost packets */
		        for(i = 6; i >=0; i--) {
				lost_packet_in_intervall[i+1]=lost_packet_in_intervall[i];
			}
			lost_packet_in_intervall[0] = mseq_offs-high_offs - 1;
	                
			log_loss(arrival);
			lastloss = arrival;
		} else { /* new lost packet in old round trip time */
			lost_packet_in_intervall[0]+=mseq_offs-high_offs - 1;
			ppl[ppl_cur] += mseq_offs-high_offs - 1;
		}
		high_offs = mseq_offs;
	}
	if (mseq_offs == high_offs + 1) { /* prevent reordered packets from changing high_offs */
/*		packet in order */
		high_offs = mseq_offs;
	}
	++ppl[ppl_cur]; /* add packet to last interval */

	return 1;
}

u_int32 calc_tzero(double rtt) {
        static int t_srtt_ = 8;
        static int t_rttvar_ = 0;
        double t_rtxcur_;
        int t_rtt_;

	t_rtt_ = (int) (rtt / MILLISEC / tcp_tick_ + 0.5);

        if (t_rtt_ < 1)
                t_rtt_ = 1;

        /*
          srtt has 3 bits to the right of the binary point
          rttvar has 2
          */
        if (t_srtt_ != 0) {
                register short delta;
                delta = t_rtt_ - (t_srtt_ >> T_SRTT_BITS);
                if ((t_srtt_ += delta) <= 0)
                        t_srtt_ = 1;
                if (delta < 0)
                        delta = -delta;
                delta -= (t_rttvar_ >> T_RTTVAR_BITS);
                if ((t_rttvar_ += delta) <= 0)
                        t_rttvar_ = 1;
        } else {
                t_srtt_ = t_rtt_ << T_SRTT_BITS;
                t_rttvar_ = t_rtt_ << (T_RTTVAR_BITS-1);
        }
	t_rtxcur_ = (((t_rttvar_ << (2 + (T_SRTT_BITS - T_RTTVAR_BITS))) +
                      t_srtt_)  >> T_SRTT_BITS ) * 0.1;
        if (t_rtxcur_ < 0.2)
                t_rtxcur_ = 0.2;
        return t_rtxcur_ * MILLISEC;
}

int main(int argc, char *argv[]) {
	char *addr = NULL;
	int  port = -1;
	int  ttl = 16;

	char send_buffer[REP_SIZE];
	char recv_buffer[PKT_SIZE];
	struct pkt_dtg *packet = (struct pkt_dtg *)recv_buffer;
	struct rep_dtg *rep    = (struct rep_dtg *)send_buffer;

	socket_udp *send_socket;
	socket_udp *recv_socket;

	char   *shmptr;
	int    shmid;
	struct shmid_ds buf;
	struct shared *info;
	int pid;

	int64   sim_start = 0;
	int64   arrival = 0, last_report = 0; 

	u_int32 rtt_cur = 0;
	u_int32 rtt_cur_old = 0;			/* rtt_cur of previous packet */
	u_int32 tzero = 0;
	double  rtt_ema = 0.0;
	double  l_act = 0.0;
	double  b_act = 0, b_exp = 0, b_rep = 0;

	double mulTFRC = 1;				/* number of TFRC connections */
	double rto;
	double j;

	int	datalen = 0, data_per_report = 0;
	
	u_int32	mode = M_INITRTT;
	int	packets = 0;
	u_int32	mseq;
	u_int32	expected_mseq = 0;			/* expected sequence number of next packet */

	u_int32 last_reported_mseq;
	int second_ack_sent = 0;
	char received_data[DATA_SIZE];

	int data_size;					/* size of data in packet*/
	FILE *file_pointer;

	int recvbuf = 0;				/* size of receiving buffer in bytes */

	static int64 last_change;
	double rateC;
	double mult;

	u_int32	last_packet_for_termination = 0;	/* variable to check if program has received new packets */
	int termination_control = 0;

	int	log_nr = 1;
	char	log_name[100];
	FILE	*h_file;

	int32 i;

	data_buffer * node, * tail, * head, * tmp;	/* pointer for linked list */
	head = NULL;					/* first element in the linked list */
	tail = NULL;					/* last element in the linked list */
	int packets_in_buffer = 0;
	int buffer_size = 100;

	int buffer_multiplier = 1;

	int end_flag;

	init_globals();

	if (argc < 3) {
		fprintf(stderr, "usage: %s address port [ttl] [-n lognr] [-w num. weights] [-m mulTFRC] [-b buffer size] [-x buffer multiplicator]\n", argv[0]);
		exit(1);
	} else {
		addr = argv[1];
		port = atoi(argv[2]);
		for (i = 3; i < argc; ++i) {
			if (argv[i][0] != '-') {
				ttl = atoi(argv[i]);
			} else {
				switch (argv[i][1]) {
				case 'n' : if (argc > i + 1) log_nr = atoi(argv[++i]); break;
				case 'w' : if (argc > i + 1) weights = atoi(argv[++i]); break;
				case 'm' : if (argc > i + 1) mulTFRC = atof(argv[++i]); break;
				case 'b' : if (argc > i + 1) buffer_size = atoi(argv[++i]); break;
				case 'x' : if (argc > i + 1) buffer_multiplier = atoi(argv[++i]); break;
				case 'c' : const_weights = 1; break;
				default : fprintf(stderr, "unknown option: %s\n", argv[i]); exit(1); break;
				}
			}
		}
	}
	set_weights();

#ifndef QUIET
	printf("started ");
	for (i = 0; i < argc; ++i)
		printf("%s ", argv[i]);
	printf("\n");
#endif

	shmid = shmget (IPC_PRIVATE, sizeof(struct shared), SHM_R | SHM_W) ; // allocates a shared memory segment 
	if (shmid == -1) {
		perror("shmget");
		exit(1);
	}
	shmptr = shmat(shmid,0,0);
	if (shmptr == NULL) {
		perror("shmat");
		exit(1);
	}
	info = (struct shared *) shmptr; // shared struct, used for termination in case of lost communication 
	shm_init(info);
	if (shmid == -1) {
		perror("shm_init");
		exit(1);
	}
	shmctl(shmid, IPC_RMID, &buf);

	udp_fd_zero();
	send_socket = udp_init(addr, port+1, ttl); /* initiate socket for sending the receiver report, stated port number+1*/
	recv_socket = udp_init(addr, port  , ttl); /* initiate socket for receiving original packets */
	if (send_socket == NULL || recv_socket == NULL) {
		perror("open socket");
		exit(1);
	}

	if(buffer_multiplier < 1) 
		buffer_multiplier = 1;
	recvbuf = buffer_multiplier * udp_getrecvbuf(recv_socket);

	sprintf(log_name, "all%i.log", log_nr);
	h_file = fopen(log_name, "w");
	if (h_file == NULL) {
		perror("open log file");
		exit(1);
	}

	init_time();
	arrival = get_time();
	sim_start = get_time() + 1; /* init sim_start */

	pid = fork();
	if(pid == 0) {
		while(termination_control == 0) {
			wait(MAX_WAIT * MICROSEC);
			if(info->end_receiver == 1){
				// terminate program 
				kill(getpid(), SIGKILL); // kill process 
				termination_control = 1;
			}
			if(last_packet_for_termination < info->sendpkt) {
				// continue 
				last_packet_for_termination = info->sendpkt;
			} else {
				// terminate program 
				if(kill(getppid(), SIGKILL) < 0)
					printf("Error\n");
				printf("Timeout because no packet has arrived in %u seconds\n", MAX_WAIT);
				termination_control = 1;
				fclose(h_file);
			} 
		}
	} else {
		while (info->end_receiver==0) { /* receiving loop */ 
			udp_fd_set(recv_socket);               

			if (udp_fd_isset(recv_socket)) {
				datalen = udp_recv(recv_socket, recv_buffer, PKT_SIZE);
				data_per_report += datalen;

				++packets;
				arrival = get_time();
				mseq    = ntohl(packet->mseq); /* get sequence number of packet */
				b_act   = ntohl(packet->bitrate);
				rtt_cur = ntohl(packet->rtt);
				rto	= ntohl(packet->rto);
				data_size = ntohl(packet->data_size);	
				end_flag = ntohl(packet->end_flag);
				if(packet->end_flag == 1) {
					end_flag = 1;
				} else {
					end_flag = 0;
				}		
				mode = ntohl(packet->mode); /* wait until RTT decreases -> M_NORMAL */

				if (rtt_ema > 0.0)
					rtt_ema = (1 - RTT_DECAY) * rtt_ema + RTT_DECAY * rtt_cur;
				else
					rtt_ema = rtt_cur;

	#ifdef NOTDEF
				/* analyzation of inc RTTs no longer needed with inter packet space modulation */
				if (rtt_cur > rtt_cur_old && mode == M_NORMAL) {
					++num_rttinc;
					if (num_rttinc >= BUF_INDICATOR) {
						mode = M_NO_INCREASE;
						mni_ts = arrival;
						b_exp_max = b_exp;
					}
				} else if (rtt_cur < rtt_cur_old) {
					if (mode == M_NO_INCREASE) {
						mode = M_NORMAL;
					}
					num_rttinc = 0;
				}
				/* timeout M_NO_INCREASE after 10 RTTs */
				if (mode == M_NO_INCREASE && (arrival - mni_ts) / MS_TO_US > 10 * rtt_ema) {
					mode = M_NORMAL;
					num_rttinc = 0;
				}
				#endif

				//tzero = max(4 * max(rtt_ema, rtt_cur), 500);
				tzero=max(calc_tzero(rtt_ema), 200);

				rtt_cur_old = rtt_cur;

				/* adjust round number only if sender has responded to the first loss event
				        * and kicked out of SLOWSTART
				        */
				if (!(mode == M_SLOWSTART && num_loss > 0))
					round = ntohl(packet->round);

				info->sendpkt=ntohl(packet->packet_no);
				update_history(info->sendpkt, arrival, rtt_cur);

				int ret;

				b_rep = data_per_report / ((double) (arrival - last_report) / MICROSEC);
				data_per_report = 0;

				/* calculate expected bitrate */
				if (num_loss > 0) {
					l_act = lossfraction();
					if(mulTFRC == 1) {
						b_exp = bitrate_eq(l_act, rtt_ema, tzero);
					}
					else {
						j = weight_lost_packets();
						rateC = mulTFRC_equation(mulTFRC, l_act, 1, (double)rtt_ema / MILLISEC, (double)tzero / MILLISEC, j);
						if(rateC > b_exp ){
							mult = (get_time()-last_change)/(rtt_ema*1000 ) ;
	        					if (mult > 2) mult = 2 ;

	        					b_exp  = b_exp  + (PKT_SIZE/((double)rtt_ema / MILLISEC))*mult ;
							if(b_exp > rateC) {
								b_exp = rateC;
							} 
						} else {
							b_exp=rateC;
						}
						last_change=get_time();
					}
				} else {
					l_act = 0.0;
					b_exp = 0.0;
				}

	#ifdef NOTDEF
				/* not used, see above */
				/* only increase above b_exp_max if not in M_NO_INCREASE mode */
				if (mode == M_NO_INCREASE && b_exp > b_exp_max) {
					b_exp = b_exp_max;
					if (b_exp - b_exp_max > 10)
						printf("%f: M_NO_INCREASE mode, bitrate inc. prevented (%2.1f / %2.1f)\n", (double) arrival / MICROSEC, b_exp, b_exp_max);
				}
	#endif
					
				if (mseq < expected_mseq) { /* retransmission */
					if(packet->start_flag==1) {
						/* resetting everything */
						printf("New StartFlag received, new Filepointer created\n");
						fclose(file_pointer);
						strncpy(received_data, packet->data, data_size);
						for(i=data_size; i<DATA_SIZE; i++) {
							received_data[i]=NULL;
						}
						file_pointer=fopen(received_data,"wb");
						second_ack_sent = 0;
						rep->mseq = packet->mseq;
						expected_mseq = mseq + 1;
						high_offs = mseq;
						if(ntohl(packet->mulTFRC)>0)
							mulTFRC = (double)ntohl(packet->mulTFRC);
						if(ntohl(packet->buffer_size)>0)
							buffer_size = (double)ntohl(packet->buffer_size);
						/* clear buffer for new transmission */
						while(head != tail) {
							tmp = head;
							head = head->next;
							free(tmp);
						}
						if(head != NULL) {
							free(head);
						}
						head = NULL;
						tail = NULL;
						packets_in_buffer = 0;
					}
					else {
						rep->mseq = htonl((u_int32)expected_mseq - 1);
						second_ack_sent++;
					}
				}
				if (mseq == expected_mseq) { /* new in order packet */
					second_ack_sent = 0;

					if(packet->start_flag==1) {
						printf("StartFlag received\n");
						strncpy(received_data, packet->data, data_size);
						for(i=data_size; i<DATA_SIZE; i++) {
							received_data[i]=NULL;
						}
						file_pointer=fopen(received_data,"wb");
						if(ntohl(packet->mulTFRC)>0)
							mulTFRC = (double)ntohl(packet->mulTFRC);
						if(ntohl(packet->buffer_size)>0) {
							buffer_size = (double)ntohl(packet->buffer_size);
							printf("Buffer: %i\n", buffer_size);
						}
						printf("Simulating %f TFRC-connections\n", mulTFRC);
					} else {
						memcpy(received_data, packet->data, data_size); 
						fwrite(received_data, sizeof(char), data_size, file_pointer);
					}
					if(packet->end_flag==1) {
						printf("EndFlag received\n");
						info->end_receiver=1;
					}
					expected_mseq++;

					/* check if expected_mseq is already in buffer */
					while(packets_in_buffer > 0) {
						if(expected_mseq == head->packet_id) {
							if(head->end_flag==1) {
								char received_data_for_last_file[head->data_size];
								memcpy(received_data_for_last_file, head->data, head->data_size); 
								printf("EndFlag received\n");
								info->end_receiver=1;
								fwrite(received_data_for_last_file, sizeof(char), head->data_size, file_pointer);
							} else {
								memcpy(received_data, head->data, head->data_size); 
								fwrite(received_data, sizeof(char), data_size, file_pointer);
							}

							if(head == tail) {
								tmp = head;
								head = NULL;
								tail = NULL;
								free(tmp);
							} else {
								tmp = head->next;
								free(head);
								head = tmp;
							}	
							expected_mseq++;
							packets_in_buffer--;
						}
						else {
							break;
						}
					}
					rep->mseq = htonl(expected_mseq - 1);

				}
					
				if (mseq > expected_mseq) { /* missing packet */
					last_reported_mseq = expected_mseq-1;
					rep->mseq = htonl((u_int32) last_reported_mseq);
					second_ack_sent++; /* send only one duplicate ack*/

					if(packets_in_buffer < buffer_size) {
						/* insert new data into the buffer */
						/* addition of new node into the linked list*/
						node = (data_buffer *)malloc(sizeof(data_buffer));
						node->data_size = data_size;
						node->end_flag = end_flag;
						if(node->end_flag == 1) 
							printf("EndFlag stored\n");
						memcpy(node->data, packet->data, data_size); 
						node->packet_id = mseq;
						if(head == NULL) {
							head = node;
							tail = node;
							packets_in_buffer++;
						} else {
							if(node->packet_id > tail->packet_id) {
								tail->next = node;
								tail = node;
								packets_in_buffer++;
							} else {
								if(node->packet_id < head->packet_id) {
									node->next = head;
									head = node;
									packets_in_buffer++;
								} else {
									tmp=head;
									while(node->packet_id > tmp->packet_id) {
										if (node->packet_id < tmp->next->packet_id) {
											node->next=tmp->next;
											tmp->next=node;
											packets_in_buffer++;
											break;
										} else {
											tmp = tmp->next;
										}
									} /* end loop for packet insert */
								} /* end head check */
							} /* end tail check*/
						} /* end head check */
					} /* end buffer_size check */
				}

				rep->dseq     = packet->dseq;
				rep->ts       = packet->ts;
				rep->b_rep    = htonl((u_int32) b_rep);
				rep->b_exp    = htonl(b_exp);
				rep->rto	= htonl(tzero);
				rep->timestamp 	= packet->timestamp;
				rep->ignore_packet = htonl(0);
				rep->recvbuf = htonl(recvbuf);
					
				if(second_ack_sent >= 2) {
					rep->ignore_packet = htonl(1);
				}
				
				while ((ret = udp_send(send_socket, (char *) rep, REP_SIZE)) < 0) {
					fprintf(stderr, "%3.2f ", (float) get_time() / MICROSEC);
					perror("udp_send");
				}

				last_report = arrival;
				packets = 0;

				/* log relevant variables */
				/* 1 time, 2 seqno, 3 round, 4 loss, 5 bact, 6 bexp, 7 brep, 8 rtt_ema, 9 rtt_cur, 10 tzero, ... */
				//fprintf(h_file, "%f %li %lu %f %1.0f %1.0f %1.0f %f %lu %lu",(float) (arrival - sim_start) / MICROSEC, mseq, round, l_act, b_act, b_exp, b_rep, rtt_ema, rtt_cur, tzero);
				/* 11 ppl_start, 12 ppl_end */
				//fprintf(h_file, " %li %li", ppl_start, ppl_end);
				//for (i = 0; i <= weights; ++i)
				//	fprintf(h_file, " %i", ppl[(ppl_cur - i + weights + 1) % (weights + 1)]); /* speed test */ 

				fprintf(h_file, "%li %li ", mseq, expected_mseq);	
				if(head != NULL) {
					tmp = head;
					while (tmp->packet_id < tail->packet_id) {
						fprintf(h_file, "%li ", tmp->packet_id);
						tmp = tmp->next;
					}
					fprintf(h_file, "%li ", tail->packet_id);
				} else {
					fprintf(h_file, "empty");
				}
	 			fprintf(h_file, "\n"); 
			} /* end packet processing */
		} /* end receiving loop */
		fclose(h_file);
		if( info->sendpkt > 0 )
			fclose(file_pointer); 
	} // end receiving process 
	shmdt(shmptr);
	return 0;
}
